-- Special thanks to Wisztom (BossStager.lua in Raid on the Spanish Armada), MaximTGMX and Zax37 (for their tips)

--[[ Inspired by Ninja Gaiden's boss Jaquio ]]--

local function ShakeScreen(mseconds)
	ffi.cast("int (*__cdecl)(int)", 0x48A000)(mseconds*1000)
end

local function NewRespawnPoint(posx, posy)
    snRes(posx,11,17)
    snRes(posy,11,18)
end

local function round(x)
		if x > 0 then
			if x - math.floor(x) > 0.5 then
				return math.floor(x+1)
			else
				return math.floor(x)
			end
		elseif x < 0 then
			if math.floor(-x) - x  > 0.5 then
				return -math.floor(math.abs(x)+1)
			else
				return -math.floor(math.abs(x))
			end
		else return 0 end
end

local function CameraCenter()
    return {X = Game(9,23,33), Y = Game(9,23,34)}
end
	
------------------------------------------- SOUND & ANIMATION -------------------------------------------
	
local function CycleLaugh(self)
    -- Sound:
	if not SoundDelay then 
		SoundDelay = GetTime()
		PlaySound("CUSTOM_ARCANIS_LAUGH")
	end
	if SoundDelay then 
		if GetTime() > SoundDelay + 2700 then
			PlaySound("CUSTOM_ARCANIS_LAUGH")
			SoundDelay = GetTime()
		end
	end
    -- Animation:
	if self.I == 2 and not Frame then
		AniTime = GetTime()
		Frame = true
	end
	if Frame and GetTime() > AniTime + 100 then
		self:SetAnimation("CUSTOM_ARCANIS_LAUGH")
		Frame = nil
	end
end


function ProjectileSound()
	if math.random(2) == 1 then return "CUSTOM_ARCANIS_ATTACK1"
	else return "CUSTOM_ARCANIS_ATTACK2" end
end


function HitSound()
	local randomsound = math.random(4)
	if randomsound == 1 then return "CUSTOM_ARCANIS_HIT1"
	elseif randomsound == 2 then return "CUSTOM_ARCANIS_HIT2"
	elseif randomsound == 3 then return "CUSTOM_ARCANIS_HIT3"
	else return "CUSTOM_ARCANIS_HIT4" end
end

------------------------------------------- MOVEMENT PATTERNS -------------------------------------------

function Bounce(self) -- State 2, State 3 and State 6 movement
	if self.Y > self.YMax or self.Y < self.YMin then
		self.SpeedY = -1*self.SpeedY
		PlaySound("CUSTOM_ARCANIS_FLY1")
	end
	if self.X > self.XMax or self.X < self.XMin then
		self.SpeedX = -1*self.SpeedX
		PlaySound("CUSTOM_ARCANIS_FLY2")
	end
end


local function Escape(self) -- State 4 movement
	if self.X > GetClaw().X + 400 then -- prevent boss from escaping too far
		self.SpeedX = 0
	end
	if self.X < GetClaw().X - 400 then -- prevent boss from being left behind
		self.SpeedX = 5
	end
	if self.X <= GetClaw().X + 400 and self.X > GetClaw().X - 400 and self.X <= self.XMax - 1600 then
		self.SpeedX = self.SpeedXChanger
	end
	Bounce(self)
end


local function Follow(self) -- State 5 movement
	if self.Y > GetClaw().Y + 160 then
		self.SpeedY = -4
	end
	if self.Y < GetClaw().Y - 160 then
		self.SpeedY = 4
	end
	Bounce(self)
end

-------------------------------------------- ATTACK PROJECTILES --------------------------------------------

function CreateProjectile(self, number) -- State 2 and State 3 attack
	local rdirx = math.random(2)
	local rdiry = math.random(2)
	if self.DetectorHook.Difficulty == 2 then 
		var = math.random(12)
	elseif self.DetectorHook.Difficulty == 1 then
		var = math.random(6)
	else
		var = math.random(9)
	end
	if rdirx == 1 then x = 1 
	else x = -1 end
	if rdiry == 1 then y = 1 
	else y = -1 end
	if var == 1 then Vx = x*1 Vy = y*1
	elseif var == 2 then Vx = x*1 Vy = y*2
	elseif var == 3 then Vx = x*2 Vy = y*1
	elseif var == 4 then Vx = x*2 Vy = y*2
	elseif var == 5 then Vx = x*3 Vy = y*1 
	elseif var == 6 then Vx = x*1 Vy = y*3
	elseif var == 7 then Vx = x*3 Vy = y*2
	elseif var == 8 then Vx = x*2 Vy = y*3
	elseif var == 9 then Vx = x*3 Vy = y*3
	elseif var == 10 then Vx = x*4 Vy = y*3
	elseif var == 11 then Vx = x*3 Vy = y*4
	elseif var == 12 then Vx = x*4 Vy = y*4
	end
	if dirx == 1 then
		table.insert(self.Projectiles, CreateObject{x=self.X+7, y=self.Y+24, z=4000, logic="CustomLogic", name="ArcanisProjectile", SpeedX=Vx, SpeedY=Vy, Damage=self.Damage})
	else
		table.insert(self.Projectiles, CreateObject{x=self.X-7, y=self.Y+24, z=4000, logic="CustomLogic", name="ArcanisProjectile", SpeedX=Vx, SpeedY=Vy, Damage=self.Damage})
	end
	if number == 2 then 
		if dirx == 1 then
			table.insert(self.Projectiles, CreateObject{x=self.X-7, y=self.Y+24, z=4000, logic="CustomLogic", name="ArcanisProjectile", SpeedX=-Vx, SpeedY=Vy, Damage=self.Damage})
		else
			table.insert(self.Projectiles, CreateObject{x=self.X+7, y=self.Y+24, z=4000, logic="CustomLogic", name="ArcanisProjectile", SpeedX=-Vx, SpeedY=Vy, Damage=self.Damage})
		end
	end
    if self.X > Game(9,23,16) - 400 and self.Y > Game(9,23,17) - 300 and self.X < Game(9,23,18) + 400 and self.Y < Game(9,23,19) + 300  then
	    PlaySound(ProjectileSound())
    end
end


function CreateSniperProjectile(self, velocity, offsetY, offsetX, MaxX) -- State 4 and State 5 attack
	self.mouth = self.Y + 32
	local speed = velocity + math.random(10)
	if self.X ~= GetClaw().X and mouth ~= GetClaw().Y then
		arctg = math.atan(math.abs(GetClaw().Y + offsetY - self.mouth)/math.abs(GetClaw().X + offsetX - self.X)) -- angle in radians
	end
	if self.X < GetClaw().X and self.Y < GetClaw().Y then -- from up-left
		Vx = round(speed*math.cos(arctg))
		Vy = round(speed*math.sin(arctg))
	elseif self.X == GetClaw().X and self.Y < GetClaw().Y then -- from up
		Vx = 0
		Vy = speed
	elseif self.X > GetClaw().X and self.Y < GetClaw().Y then -- from up-right
		Vx = -round(speed*math.cos(arctg))
		Vy = round(speed*math.sin(arctg))
	elseif self.X > GetClaw().X and self.Y == GetClaw().Y then -- from right
		Vx = -speed
		Vy = 0
	elseif self.X > GetClaw().X and self.Y > GetClaw().Y then -- from down-right
		Vx = -round(speed*math.cos(arctg))
		Vy = -round(speed*math.sin(arctg))
	elseif self.X == GetClaw().X and self.Y > GetClaw().Y then -- from down
		Vx = 0
		Vy = -speed
	elseif self.X < GetClaw().X and self.Y > GetClaw().Y then -- from down-left
		Vx = round(speed*math.cos(arctg))
		Vy = -round(speed*math.sin(arctg))
	else													-- from left
		Vx = speed
		Vy = 0
	end
	CreateObject{x=self.X, y=self.mouth, z=4000, logic="CustomLogic", name="ArcanisProjectile2", SpeedX=Vx , SpeedY=Vy, XMax=MaxX, Damage = self.Damage}
	if self.X > Game(9,23,16) - 400 and self.Y > Game(9,23,17) - 300 and self.X < Game(9,23,18) + 400 and self.Y < Game(9,23,19) + 300  then
	    PlaySound(ProjectileSound())
    end
end


local function CreateRandomProjectile(self) -- State 6 attack
	local Random = math.random(2)
	if Random == 1 then 
		local RandomVelocity = math.random(20,40)
		local RandomOffsetX = math.random(90) - math.random(90)
		local RandomOffsetY = math.random(90) - math.random(90)
		CreateSniperProjectile(self, RandomVelocity, RandomOffsetX, RandomOffsetY, 0)
	else 
		CreateProjectile(self, 2) 
	end
end

-------------------------------------------- HUD ELEMENTS --------------------------------------------

local function CreateHUDElement(posX, posY, objImage)
    local flags = ffi.new("Flags_t")
    flags.AlwaysActive = true
    return CreateObject{x=posX,y=posY,z=11000,logic="BackgroundLogic",image=objImage, Flags=flags}
end


local function UpdateBossBar(self)
	self.BossBar.I = math.floor(40*self.Health/self.StartingHealth + 1)
	self.BossBar:SetFrame(self.BossBar.I)
end

------------------------------------------------- MAIN -------------------------------------------------

function init(self)
    self.XMin = self.X - 550
    self.XMax = self.X + 491
    self.Ymin = self.Y - 283
    self.YMax = self.Y + 193
end


function main(self)
	
    self.DetectorHook = GetObject(3206)

    -- Screen coordinates:
    PLAY_AREA = ffi.cast("Rect*", 0x535840)[0]
	local CenterX = PLAY_AREA.Right/2 + 1
	local CenterY = PLAY_AREA.Bottom/2 + 1
	local CenterBossName = (PLAY_AREA.Bottom+1)*0.3-1
	local EndBossName = PLAY_AREA.Right + 138
	local BossNameStartY = -66
    local BossBarY = ((PLAY_AREA.Bottom+1)*0.9+15)
	

	-- Initialisation:
	if self.State == 0 then

		-- important stuff:
		self:SetFrame(101)
		self.AttackTypeFlags, self.ObjectTypeFlags, self.HitTypeFlags = 2, 4, 2
		self.AttackRect, self.HitRect = {-30,-60,30,32}, {-25,-55,25,25}
		if self.DetectorHook.Difficulty == 2 then 
			self.Damage = 10
		elseif self.DetectorHook.Difficulty == 1 then 
			self.Damage = 2
		else
			self.Damage = 5
		end
        self.Health = 200
        self.StartingHealth = 200
        -- projectiles table:
        self.Projectiles = { }

		-- initialise other variables:
        self.StartingX, self.StartingY = self.X, self.Y
		self.StagerNo, self.FinalPhaseNo, self.EndingNo = 0, 0, 0
		self.DamageTimeDelay, self.AttackTimeDelay = 0, 0
		self.InternalTimeDelay, self.HitTimeDelay = 0, 0

        -- start:
		self.State = 1

	end

	
	 -- General instructions:
	if (self.State >= 1 and self.State <= 6) or self.State == 19 or self.State >= 103 then

        --Movement:
		self:AnimationStep()
		self.X = self.X + self.SpeedX
		self.Y = self.Y + self.SpeedY

        -- Updating the boss health bar:
		if self.BossBar then UpdateBossBar(self) end

        -- Various situations:
		if self.State ~= 19 then
			-- if boss is defeated:
			if self.Health <= 0 then 
			    self.State = 19
			end
			-- if Claw dies:
            if GetClaw().Health <= 0 then
                -- by losing health:
			    if GetClaw().I == 402 then 
				    self.State = 21
				    self.TimeCounter = GetTime() + 1100
			    --  by falling into abyss/on spikes:
			    elseif GetClaw().I == 403 then
				    self.State = 21
				    self.TimeCounter = GetTime() + 2800
                -- else (?)
                else
                    self.State = 21
				    self.TimeCounter = GetTime() + 1100
			    end
            end
			-- remove projectile's reference from the table as soon as it gets destroyed:
			for index,projectile in ipairs(self.Projectiles) do
				if projectile.IsDestroyed then
					if projectile.IsDestroyed == true then projectile = 0 end
				end
			end
		end

		-- unlock final phase when boss hp reaches 25:
		if self.Health <= 25 and not self.FinaleUnlocked then
            self.FinaleUnlocked = true
			self.Peg1 = CreateObject{x=self.StartingX + 11395, y=self.StartingY - 2423, z=1000 , logic="CustomLogic", name="ArcanisPeg"}
			self.Peg2 = CreateObject{x=self.StartingX + 11331, y=self.StartingY - 2551, z=1000 , logic="CustomLogic", name="ArcanisPeg"}
		end

	end
	

	 -- Stager:
	if self.State == 1 then
        -- freeze Claw, set respawn point:
		if self.StagerNo == 0 then 
			if GetClaw().I <= 18 then
                SetMusic("ARCANIS")
				GetClaw():SetAnimation("CLAW_STAND", 100)
				GetClaw().State = 5009
				NewRespawnPoint(GetClaw().X,GetClaw().Y)
				self.StagerNo = 1
			end
		end

        -- create the boss hp bar, set the Always Active flag:
		if self.StagerNo == 1 then 
			self.Flags.AlwaysActive = true
			self.BossBar = CreateHUDElement(CenterX, BossBarY, "CUSTOM_BOSSBAR")
			CameraToObject(self)
			self.TimeCounter = GetTime()
			self.StagerNo = 2
		end

        -- boss lights up:
		if self.StagerNo == 2 and GetTime() > self.TimeCounter + 500 then
			self:SetAnimation("CUSTOM_ARCANIS_FIREUP")
			PlaySound("CUSTOM_ARCANIS_START")
			self.TimeCounter = GetTime()
			self.StagerNo = 3
		end

        -- boss starts laughing:
		if self.StagerNo == 3 and GetTime() > self.TimeCounter + 1100 then
			self.BossExclamation = CreateObject {x=self.X, y=self.Y-90, z=9000, logic="BlinkingEyes", image="GAME_EXCLAMATION"}
			self.TimeCounter = GetTime()
			self.StagerNo = 4
		end

        -- laughing cycle:
		if self.StagerNo >= 3 and self.StagerNo <= 7 then
			CycleLaugh(self)
		end

        -- Claw speaks:
		if self.StagerNo == 4 and GetTime() > self.TimeCounter + 2200 then
			CameraToObject(GetClaw())
			ClawSound("CUSTOM_TRIGGER_NOCOMINGBACK")
			self.TimeCounter = GetTime()
			self.StagerNo = 5
		end

        -- create the boss name title:
		if self.StagerNo == 5 and GetTime() > self.TimeCounter + 2200 then 
			ShakeScreen(3)
			self.BossName = CreateHUDElement(CenterX, BossNameStartY, "CUSTOM_BOSSNAME")
			PlaySound("GAME_SDPT1")
			self.TimeCounter = GetTime()
			self.StagerNo = 6
		end

        -- boss name title animation:
		if self.StagerNo == 6 then
			local AnimationLenght = 2029
			self.BossName.Y = BossNameStartY + (CenterBossName-BossNameStartY) * (GetTime()-self.TimeCounter) / AnimationLenght
			if GetTime() > self.TimeCounter + AnimationLenght then
				self.TimeCounter = GetTime()
				self.StagerNo = 7
			end
		end

        -- play a sound and stop shaking:
		if self.StagerNo == 7 then
			PlaySound("GAME_SDPT2")
			CameraToClaw()
			self.TimeCounter = GetTime()
			self.BossExclamation:Destroy()
			self.StagerNo = 8
		end

        -- wait for the sound to stop:
		if self.StagerNo == 8 and GetTime() > self.TimeCounter + 800 then
			self.TimeCounter = GetTime()
			self.StagerNo = 9
		end

        -- boss name title slides out:
		if self.StagerNo == 9 then
			local AnimationLenght = 750
			self.BossName.X = CenterX + (EndBossName-CenterX) * (GetTime()-self.TimeCounter) / AnimationLenght
			if GetTime() > self.TimeCounter + AnimationLenght then
				self.StagerNo = 10
			end
		end

        -- last preparations:
        if self.StagerNo == 10 then
            GetClaw().State = 24 -- unfreeze Claw
            self:SetFrame(103)
            self.BossName:Destroy()
            self.StagerNo = 11
        end

        -- the fight starts:
		if self.StagerNo == 11 then
			self.SpeedX, self.SpeedY = -2, 2
			self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")
			self.State = 2
		end

	end
	

	-- Phase 1, calm:
	if self.State == 2 then 
		-- movement pattern:
		Bounce(self)
		-- attack pattern with sound and animation:
		if GetTime() > self.AttackTimeDelay and self.I == 6 then
			CreateProjectile(self, 1)
			self.AttackTimeDelay = GetTime() + 300
		elseif GetTime() > self.AttackTimeDelay + 50 and self.I == 2 then
			self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")
			self.AttackTimeDelay = GetTime() + 300
		end
		-- go to the next phase when boss hp reaches 3/4 of his starting hp:
		if self.Health <= 3*self.StartingHealth/4 then
			self.State = 103
		end
	end


    -- Prepare for phase 2:
    if self.State == 103 then
        self.JustKeepCalm1 = CreateObject{x=self.StartingX - 448, y=self.StartingY + 32, z=1010, logic="CustomLogic", name="ArcanisHealthPotion"}
		self.JustKeepCalm2 = CreateObject{x=self.StartingX + 384, y=self.StartingY + 32, z=1010, logic="CustomLogic", name="ArcanisHealthPotion"}
		self:SetAnimation("CUSTOM_ARCANIS_ATTACK2")
        self.State = 3
    end
	

	-- Phase 2, double ice balls:
	if self.State == 3 then 
		-- movement pattern:
		Bounce(self)
		-- attack pattern with sound and animation:
		if GetTime() > self.AttackTimeDelay and self.I == 6 then
			CreateProjectile(self, 2)
			self.AttackTimeDelay = GetTime() + 400
		elseif GetTime() > self.AttackTimeDelay and self.I == 2 then
			self:SetAnimation("CUSTOM_ARCANIS_ATTACK2")
			self.AttackTimeDelay = GetTime() + 400
		end
		-- go to the next phase when boss hp reaches 1/2 of his starting hp:
		if self.Health <= self.StartingHealth/2 then
			self.State = 104
		end
	end


    -- Prepare for phase 3:
    if self.State == 104 then
        -- attack pattern with sound and animation:
		if GetTime() > self.AttackTimeDelay and self.I == 6 then
			CreateProjectile(self, 2)
			self.AttackTimeDelay = GetTime() + 400
		elseif GetTime() > self.AttackTimeDelay and self.I == 2 then
			self:SetAnimation("CUSTOM_ARCANIS_ATTACK2")
			self.AttackTimeDelay = GetTime() + 400
		end
        -- escape:
        if self.Y <= self.YMax - 16 and self.X < self.XMax - 100 then
			self.SpeedY, self.SpeedX = 2, 2	
		else
			self.SpeedY, self.SpeedX = 0, 4	
			self.XMax, self.YMin = self.StartingX + 12800, self.StartingY - 105
			if self.X > self.StartingX + 800 then
				self.SpeedY = 2
				self.SpeedXChanger = 3
				self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")
				self.JustKeepCalm3 = CreateObject{x=self.StartingX + 1250, y=self.StartingY + 160, z=1010, logic="CustomLogic", name="ArcanisHealthPotion"}
				self.JustKeepCalm4 = CreateObject{x=self.StartingX + 11200, y=self.StartingY + 160, z=1010, logic="CustomLogic", name="ArcanisHealthPotion"}
				self.ElevatorLong = CreateObject{x=self.StartingX+560, y=self.StartingY+256, z=-5, logic="CustomLogic", name="ArcanisElevator"}
				self.State = 4
			end
		end
    end
	

	-- 3rd phase ("Escape"):
	if self.State == 4 then 
		-- movement pattern:
		Escape(self)
		if GetTime() > self.InternalTimeDelay then -- back to normal speed after short acceleration
			self.SpeedXChanger = 3 
		end
		-- attack pattern with sound and animation:
		if GetTime() > self.AttackTimeDelay then
			if not self.ShootFrame and self.I == 6 then -- sync animation with projectile shot (shoots, when the jaw is fully open)
				if GetClaw().X > self.StartingX + 750 then 	-- prevent projectile from flying through the wall
					CreateSniperProjectile(self, 40, 10, 60, 0) 
				end
				self.AttackTimeDelay = GetTime() + 300		-- prevent creating multiple projectiles
				self.ShootFrame = true						-- save that the jaw has opened
			end
			if self.ShootFrame and self.I == 2 then -- if jaw was opened and now it's closed then
				self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")	-- repeat the animation
				self.ShootFrame = false						-- save that the jaw has closed
			end
		end
		-- go to the next phase when boss reaches certain X:
		if self.X >= self.XMax - 1800 then
			self.State = 105
		end
	end


    -- Prepare for phase 4:
    if self.State == 105 then
        self.XMin, self.XMax = math.floor(self.XMax - 1920), math.floor(self.XMax - 1680)
		self.YMin, self.YMax = math.floor(self.YMin - 3500), math.floor(self.YMax + 160)
		self.ElevatorLong.State = 4
		self.JustKeepCalm5 = CreateObject{x=self.StartingX+10692, y=self.StartingY-3220, z=1010, logic="CustomLogic", name="ArcanisHealthPotion"}
		self.JustKeepCalm6 = CreateObject{x=self.StartingX+10692, y=self.StartingY-3800, z=1010, logic="CustomLogic", name="ArcanisHealthPotion"}
		PlaySound("CUSTOM_ARCANIS_LAUGH2")
		self.SpeedX, self.SpeedY = 2, -3
		self:SetAnimation("CUSTOM_ARCANIS_ATTACK2")
		self.State = 5
    end
	

	-- 4th phase ("Sniper"):
	if self.State == 5 then
		-- movement pattern:
		Follow(self)
		-- attack pattern with sound and animation:
		if GetTime() > self.AttackTimeDelay and self.I == 6 then
			CreateSniperProjectile(self, 20, 20, -100, self.StartingX + 11461)
			self.AttackTimeDelay = GetTime() + 300
		elseif GetTime() > self.AttackTimeDelay and self.I == 2 then
			self:SetAnimation("CUSTOM_ARCANIS_ATTACK2")
			self.AttackTimeDelay = GetTime() + 400
		end
		-- go to the final phase when boss hp reaches 25:
		if self.Health <= 25 then
			self.State = 106
		end
	end


     -- Prepare for phase 5:
    if self.State == 106 then
        self.SpeedX, self.SpeedY = 0, -4
		self.XMin= self.XMin - 500
		self.YMin, self.YMax = self.StartingY - 2903, self.StartingY - 2703
		self.FinalPhaseNo = 1
        self.State = 6
    end

	
	-- 5th phase ("Berserk"):
	if self.State == 6 then

		-- berserk mode:
		if self.FinalPhaseNo % 3 == 2 then
			-- movement pattern:
			Bounce(self)
			-- attack pattern with sound and animation:
			if GetTime() > self.AttackTimeDelay then
				if not self.ShootFrame and self.I == 6 then
					CreateRandomProjectile(self) 
					self.AttackTimeDelay = GetTime() + 300
					self.ShootFrame = true
				end
				if self.ShootFrame and self.I == 2 then
					self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")
					self.ShootFrame = false
				end
			end

		-- transitional sub-phases:	
		elseif self.FinalPhaseNo % 3 == 1 and self.Y < self.YMax then
			PlaySound("CUSTOM_ARCANIS_QUAKE")
			ShakeScreen(3)
			self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")
			self.SpeedY = -2
			if self.X < (self.XMin+self.XMax)/2 then 
                self.SpeedX = 2 
            else 
                self.SpeedX = -2 
            end
			self.FinalPhaseNo = self.FinalPhaseNo + 1

        elseif self.FinalPhaseNo % 3 == 0 then
            self.SpeedX, self.SpeedY = 0, -4
            local multiplier = self.FinalPhaseNo/3 - 1
            self.YMin = self.StartingY - 3350 - 360*multiplier
            self.YMax = self.StartingY - 3150 - 360*multiplier
            self.FinalPhaseNo = self.FinalPhaseNo + 1

		end

	end
	

	-- Boss dies:
	if self.State == 19 then 

		if self.EndingNo == 0 then
            SetMusic("SILENCE")
			self.SpeedX, self.SpeedY = 0, 0 -- freeze Boss
			self.LastClawsState = GetClaw().State -- save last Claw's state
			GetClaw().State = 5009
            NewRespawnPoint(self.XTweak, self.YTweak + 100)
			self.TimeCounter = GetTime()
			self.EndingNo = 1
		end

        if self.EndingNo == 1 then
            PlaySound("CUSTOM_ARCANIS_DIE")
            CameraToObject(self)
            -- freeze all the projectiles:
			for _, projectile in ipairs(self.Projectiles) do
		        if projectile.IsDestroyed == false then projectile.State = 28 end
			end 
            self.EndingNo = 2
        end

		if self.EndingNo == 2 and GetTime() > self.TimeCounter + 2200 then -- wait for a sound to stop
			local BrokenSkull = CreateObject{x=self.X, y=self.Y, z=self.Z+1, logic="DoNothingNormal", image="CUSTOM_ARCANIS"}
            BrokenSkull:SetFrame(201)
			self.DrawFlags.NoDraw = true -- make defeated boss invisible for a while
			self.TimeCounter = GetTime()
			self.EndingNo = 3
		end

		if self.EndingNo == 3 and GetTime() > self.TimeCounter + 400 then -- broken skull falls down
			self.MapPiece = CreateHUDElement(CenterX, CenterY, "GAME_MAPPIECE") -- create temporary Map Piece HUD
			CameraToPoint(self.XTweak, self.YTweak) -- camera to Map Piece destination
			self.TimeCounter = GetTime()
			PlaySound("GAME_AMULETRISE") 
			self.EndingNo = 4
		end

		if self.EndingNo == 4 and GetTime() > self.TimeCounter + 2000 then -- wait for a sound to stop
			if (CameraCenter().X == self.XTweak and CameraCenter().Y == self.YTweak) or GetTime() > self.TimeCounter + 6000 then -- wait until the camera reaches the point and create the real Map Piece there
				CreateObject{x=self.XTweak, y=self.YTweak, z=2000, logic="SpecialPowerup", image="GAME_MAPPIECE"}
				self.MapPiece:Destroy()
				self.EndingNo = 5
			end
		end

		if self.EndingNo == 5 and GetTime() > self.TimeCounter + 6000 then
			CameraToClaw() -- camera back to Claw
			GetClaw().State = self.LastClawsState -- unfreeze Claw
			self.State = 20
		end

	end
	

	 -- Boss is dead:
	if self.State == 20 then
		self:Destroy()
	end
	

	-- Claw dies, reset the fight:
	if self.State == 21 and GetTime() > self.TimeCounter then 
		self.Health = self.StartingHealth
		self.SpeedX, self.SpeedY = -2, 2
		self.XMin, self.XMax = self.StartingX - 550, self.StartingX + 491
		self.YMin, self.YMax = self.StartingY - 283, self.StartingY + 193
		if self.BossExclamation then self.BossExclamation:Destroy() end
		if self.BossName then self.BossName:Destroy() end
		if self.ElevatorLong then self.ElevatorLong:Destroy() end
		if self.JustKeepCalm1 then self.JustKeepCalm1:Destroy() end
		if self.JustKeepCalm2 then self.JustKeepCalm2:Destroy() end
		if self.JustKeepCalm3 then self.JustKeepCalm3:Destroy() end
		if self.JustKeepCalm4 then self.JustKeepCalm4:Destroy() end
		if self.JustKeepCalm5 then self.JustKeepCalm5:Destroy() end
		if self.JustKeepCalm6 then self.JustKeepCalm6:Destroy() end
		if self.ShootFrame then self.ShootFrame = false end
		if self.Peg1 then self.Peg1:Destroy() end
		if self.Peg2 then self.Peg2:Destroy() end
		if self.FinaleUnlocked then self.FinaleUnlocked = false end
		if self.FinalPhaseNo then self.FinalPhaseNo = 0 end
		for _, projectile in ipairs(self.Projectiles) do
			if projectile ~= 0 then projectile:Destroy() end
		end
		self.Projectiles = { }
		self:SetAnimation("CUSTOM_ARCANIS_ATTACK1")
        self.X, self.Y = self.StartingX, self.StartingY
		self.State = 2
	end

end


function attack(self)
	if self.State >= 1 and self.State <= 6 then
		if GetTime() > self.DamageTimeDelay then -- the skull itself can hit Claw max 3 times in 1 sec
			self.DealDmg = {x=GetClaw().X, y=GetClaw().Y, z=0, logic="OmarBullet", image="GAME_SOUNDICON", Damage=1}
			self.DealDmg.AttackRect = {-16,-16,16,16}
			self.DamageTimeDelay = GetTime() + 333
		end
	end
end


function hit(self)
	if GetTime() >= self.HitTimeDelay then

		if (self.State >= 1 and self.State <= 6) or self.State > 102 then 
			if GetClaw().Damage >= 20 then -- set a limit for the damage the boss can take
				self.Health = math.floor(self.Health - 16)
			else
				self.Health = math.floor(self.Health - GetClaw().Damage)
			end
			CreateObject {x=self.X, y=self.Y, z=5000, logic="CustomLogic", name="ArcanisTakesHit", image="GAME_ENEMYHIT"}
			PlaySound(HitSound())
			PlaySound("GAME_SOLHITHI")
			if self.State ~= 6 then -- Claw can hit the skull max 3 times in 1 sec (except for final phase)
				self.HitTimeDelay = GetTime() + 333
			end
		end

		-- short acceleration in the "Escape" phase:
		if self.State == 4 then 
			self.SpeedXChanger = 5
			self.InternalTimeDelay = GetTime() + 800
		end	

		-- breaking berserk mode:
		if self.State == 6 and GetTime() > self.InternalTimeDelay and self.FinalPhaseNo % 3 == 2 then
			self.FinalPhaseNo = self.FinalPhaseNo + 1
			self.InternalTimeDelay = GetTime() + 1000
			self.HitTimeDelay = self.InternalTimeDelay
		end

	end
end
