--------------------------------------------------------------------------------------- PEGLEG MODULE ---------------------------------------------------------------------------------------
pl = { 

    TileType = {
        Single = 0x51414C, 
        Double = 0x51418C
    },

    Attribute = {
        Clear = 0, 
        Solid = 1, 
        Ground = 2, 
        Climb = 3,  
        Death = 4
    }
    
}
------------------------------------------------------------------------------------------ PLANES ------------------------------------------------------------------------------------------

function pl:DrawPlane(plane_index)
    if tonumber(plane_index) >= 0 and tonumber(plane_index) < pl:PlanesCount() then
        local _flags = Game(9,14,tonumber(plane_index),2)
        snRes(bit.bxor(2,_flags),12,9,14,tonumber(plane_index),2) 
    else
        MessageBox("Error: no plane with index " .. plane_index .."!\nThe last plane has index " .. pl:FrontPlane())
    end
end

function pl:FrontPlane()
    return Game(9,15) - 1
end

function pl:GetPlane(plane_index, property)
    if tonumber(plane_index) >= 0 and tonumber(plane_index) < pl:PlanesCount() then
        local _plane = {
                FLAGS = Game(9,14,tonumber(plane_index),2),
                WIDTH = Game(9,14,tonumber(plane_index),10),
                HEIGHT = Game(9,14,tonumber(plane_index),11),
                TILEWIDTH = Game(9,14,tonumber(plane_index),14),
                TILEHEIGHT = Game(9,14,tonumber(plane_index),15),
                Z = Game(9,14,tonumber(plane_index),32),
                MOVEX = Game(9,14,tonumber(plane_index),37),
                MOVEY = Game(9,14,tonumber(plane_index),38),
                FILL = Game(9,14,tonumber(plane_index),81)
            }
        if _plane[string.upper(property)] then
            return _plane[string.upper(property)]
        else
            MessageBox([["Error: no plane property named " .. property .. "!\n"
            'Available are "Flags", "Width", "Height", "TileWidth", "TileHeight", "Z", "MoveX", "MoveY" and "Fill".']])
        end
    else
        MessageBox("Error: no plane with index " .. plane_index .."!\nThe last plane has index " .. pl:FrontPlane())
    end
end

function pl:MainPlane()
    return Game(9,23,1)
end

function pl:PlanesCount()
    return Game(9,15)
end

function pl:SetPlaneZ(plane_index, z)
    if tonumber(plane_index) >= 0 and tonumber(plane_index) < pl:PlanesCount() then
        snRes(z,12,9,14,tonumber(plane_index),32)
    else
        MessageBox("Error: no plane with index " .. plane_index .."!\nThe last plane has index " .. pl:FrontPlane())
    end
end

function pl:SetPlaneFill(plane_index, color)
    if tonumber(plane_index) >= 0 and tonumber(plane_index) < pl:PlanesCount() then
        snRes(color%256,12,9,14,tonumber(plane_index),81)
    else
        MessageBox("Error: no plane with index " .. plane_index .."!\nThe last plane has index " .. pl:FrontPlane())
    end
end

------------------------------------------------------------------------------------------ TILES ------------------------------------------------------------------------------------------

ffi.cdef[[
    typedef struct TileA {
        int Type;
        const int TileWidth;
        const int TileHeight;
        int Attribute;
        int RectAttribute;
        Rect TileRect;
    } TileA;
]]

local _attribute = {CLEAR = 0, SOLID = 1, GROUND = 2, CLIMB = 3,  DEATH = 4}
local _tileType = {SINGLE = 0x51414C, DOUBLE = 0x51418C}

function pl:Attribute(str)
    return _attribute[string.upper(str)]
end

function pl:GetTile(plane_index, x, y)
    if tonumber(plane_index) >= 0 and tonumber(plane_index) < pl:PlanesCount() then
        if x >= 0 and x < Game(9,14,tonumber(plane_index),10) and y >= 0 and y < Game(9,14,tonumber(plane_index),11) then
            local _TileIndex = x + y * Game(9,14,tonumber(plane_index),10)
            return ffi.cast("int*",Game(9,14,tonumber(plane_index),8))[_TileIndex]
        else
            return -3
        end
    else
        MessageBox("Error: no plane with index " .. plane .."!\nThe last plane has index " .. pl:FrontPlane())
    end
end

function pl:GetTileA(id)
    return ffi.cast("TileA*", ffi.cast("int***",Game(9,19,id)))
end

function pl:PlaceTile(id, plane_index, x, y)
    if tonumber(plane_index) >= 0 and tonumber(plane_index) < pl:PlanesCount() then
        if x >= 0 and x < Game(9,14,plane_index,10) and y >= 0 and y < Game(9,14,plane_index,11) then
            local _TileIndex = x + y * Game(9,14,plane_index,10)
            ffi.cast("int*",Game(9,14,plane_index,8))[_TileIndex] = id
        else
            MessageBox("Error: coordinates out of plane boundaries!")
        end
    else
        MessageBox("Error: no plane with index " .. plane_index .."!\nThe last plane has index " .. pl:FrontPlane())
    end
end

function pl:SetTile(id, plane_index, x, y)
    pl:PlaceTile(id, plane_index, x, y)
end

function pl:SetTileA(AttTable)
    if tonumber(AttTable.ID) then
        local tile = pl:GetTileA(tonumber(AttTable.ID))
        if AttTable.Type then
            if _tileType[string.upper(AttTable.Type)] then
                tile.Type = _tileType[string.upper(AttTable.Type)]
            else
                MessageBox('Error: invalid tile type, use "Single" or "Double"')
            end
        end
        if AttTable.Attribute then
            if _attribute[string.upper(AttTable.Attribute)] then
                tile.Attribute = _attribute[string.upper(AttTable.Attribute)]
            else
                MessageBox('Error: invalid tile attribute, use "Clear", "Solid", "Ground", "Climb" or "Death".')
            end
        end
        if AttTable.RectAttribute then
            if _attribute[string.upper(AttTable.RectAttribute)] then
                tile.RectAttribute = _attribute[string.upper(AttTable.RectAttribute)]
            else
                MessageBox('Error: invalid rect tile attribute, use "Clear", "Solid", "Ground", "Climb" or "Death".')
            end
        end
        if AttTable.TileRect then
            if AttTable.TileRect[1] >= 0 and AttTable.TileRect[1] < tile.TileWidth and 
            --[[&]] AttTable.TileRect[3] >= 0 and AttTable.TileRect[3] < tile.TileWidth and
            --[[&]] AttTable.TileRect[2] >= 0 and AttTable.TileRect[2] < tile.TileHeight and 
            --[[&]] AttTable.TileRect[4] >= 0 and AttTable.TileRect[4] < tile.TileHeight then
                tile.TileRect.Left = AttTable.TileRect[1]
                tile.TileRect.Top = AttTable.TileRect[2]
                tile.TileRect.Right = AttTable.TileRect[3]
                tile.TileRect.Bottom = AttTable.TileRect[4]
            else
                MessageBox('Error: TileRect size out of the tile size (' .. tile.TileWidth .. "x" .. tile.TileHeight .. ")!")
            end
        end
    else
        MessageBox('Error: invalid tile ID!')
    end
end

function pl:TileType(str)
    return _tileType[string.upper(str)]
end

------------------------------------------------------------------------------------------ PHYSICS ------------------------------------------------------------------------------------------


local function RealTileRect(rect)
    return ffi.new("Rect", {math.min(rect.Left, rect.Right), math.min(rect.Top, rect.Bottom), math.max(rect.Left, rect.Right), math.max(rect.Top, rect.Bottom)})
end

local function MoveXPointTile(x,y,rect,dir,inside)
    local _tileRect = RealTileRect(rect)
    local distance = ffi.new("int", 0)
    local collision = ffi.new("bool", false)
    if inside == true then
        -- check if collision is possible:
        if y > _tileRect.Top and y < _tileRect.Bottom then
        -- moving in the right direction:
            if dir == 6 then
            -- case 1: past the rect:
                if x > _tileRect.Right then
                    distance, collision = 64 - x, false
            -- case 2: before the rect:
                elseif x < _tileRect.Left then
                    distance, collision =  math.abs(_tileRect.Left - x), true
            -- case 3: inside the rect:
                else
                    distance, collision =  0, true
                end
        -- moving in the left direction:
            elseif dir == 4 then
            -- case 1: past the rect:
                if x < _tileRect.Left then
                    distance, collision =  x, false
            -- case 2: before the rect:
                elseif x > _tileRect.Right then
                    distance, collision =  math.abs(x - _tileRect.Right), true
            -- case 3: inside the rect:
                else
                    distance, collision =  0, true
                end
            end
        else -- collision is not possible
            if dir == 6 then
                distance, collision =  64 - x, false
            elseif dir == 4 then
                distance, collision =  x + 1, false
            end
        end
    else
        --check if collision already happens at the start:
        if y < _tileRect.Top or y > _tileRect.Bottom then
            distance, collision =  0, true
        --if not
        else
            if dir == 6 then
            -- case 1: outside the rect
                if x < _tileRect.Left then
                    distance, collision = 0, true
            -- case 2: inside the rect, but going out of rect
                elseif _tileRect.Right < pl:GetPlane(pl:MainPlane(), "TileWidth") - 1 then
                    distance, collision =  math.abs(_tileRect.Right - x), true
            -- case 3: inside the rect to the end
                else
                    distance, collision =  64 - x, false
                end
            elseif dir == 4 then
             -- case 1: outside the rect
                if x > _tileRect.Right then
                    distance, collision = 0, true
            -- case 2: inside the rect, but going out of rect
                elseif _tileRect.Left > 0 then
                    distance, collision =  math.abs(x - _tileRect.Left), true
            -- case 3: inside the rect to the end
                else
                    distance, collision =  x + 1, false
                end    
            end
        end
    end
    return distance, collision
end

local function MoveYPointTile(x,y,rect,dir,inside)
    local _tileRect = RealTileRect(rect)
    local distance = ffi.new("int", 0)
    local collision = ffi.new("bool", false)
    if inside == true then
        -- check if collision is possible:
        if x > _tileRect.Left and x < _tileRect.Right then
        -- moving in the right direction:
            if dir == 2 then
            -- case 1: past the rect:
                if y > _tileRect.Bottom then
                    distance, collision = 64 - y, false
            -- case 2: before the rect:
                elseif y < _tileRect.Top then
                    distance, collision =  math.abs(_tileRect.Top - y), true
            -- case 3: inside the rect:
                else
                    distance, collision =  0, true
                end
        -- moving in the left direction:
            elseif dir == 8 then
            -- case 1: past the rect:
                if y < _tileRect.Top then
                    distance, collision =  y, false
            -- case 2: before the rect:
                elseif y > _tileRect.Bottom then
                    distance, collision =  math.abs(y - _tileRect.Bottom), true
            -- case 3: inside the rect:
                else
                    distance, collision =  0, true
                end
            end
        else -- collision is not possible
            if dir == 2 then
                distance, collision =  64 - y, false
            elseif dir == 8 then
                distance, collision =  y + 1, false
            end
        end
    else
        --check if collision already happens at the start:
        if x < _tileRect.Left or x > _tileRect.Right then
            distance, collision =  0, true
        --if not:
        else
            if dir == 2 then
            -- case 1: outside the rect
                if y < _tileRect.Top then
                    distance, collision = 0, true
            -- case 2: inside the rect, but going out of rect
                elseif _tileRect.Bottom < pl:GetPlane(pl:MainPlane(), "TileHeight") - 1 then
                    distance, collision =  math.abs(_tileRect.Bottom - y), true
            -- case 3: inside the rect to the end
                else
                    distance, collision =  64 - y, false
                end
            elseif dir == 8 then
             -- case 1: outside the rect
                if y > _tileRect.Bottom then
                    distance, collision = 0, true
            -- case 2: inside the rect, but going out of rect
                elseif _tileRect.Top > 0 then
                   distance, collision =  math.abs(y - _tileRect.Top), true
            -- case 3: inside the rect to the end
                else
                    distance, collision =  y + 1, false
                end    
            end
        end
    end
    return distance, collision
end

function pl:CollisionDistance(x,y,dir)
    local xTile = ffi.new("int", math.floor(x/64))
    local yTile = ffi.new("int", math.floor(y/64))
    local tile = ffi.new("TileA*");
    local distance = ffi.new("int", 0)
    local collision = ffi.new("bool", false)
    local action = pl:MainPlane()
    x, y = math.fmod(x,64), math.fmod(y,64)
    if dir == 6 then
        repeat
            if xTile >= 0 and xTile < Game(9,23,10) then
                tile = pl:GetTileA(pl:GetTile(action, xTile, yTile))
                if ((tile.Attribute == 1 or tile.Attribute == 4) and tile.Type == pl.TileType.Single) or 
                --[[|]] (tile.Type == pl.TileType.Double and (tile.RectAttribute == 1 or tile.RectAttribute == 4) and (tile.Attribute == 1 or tile.Attribute == 4)) then 
                    collision = true
                elseif tile.Type == pl.TileType.Double and (tile.RectAttribute == 1 or tile.RectAttribute == 4) then -- double inside solid or death, outside clear
                    deltaDistance, collision = MoveXPointTile(x,y,tile.TileRect,6,true)
                    distance = distance + deltaDistance
                elseif tile.Type == pl.TileType.Double and (tile.Attribute == 1 or tile.Attribute == 4) then -- double outside solid or death
                    deltaDistance, collision = MoveXPointTile(x,y,tile.TileRect,6,false)
                    distance = distance + deltaDistance
                else -- all the other tiles
                    distance = distance + 64 - math.fmod(x,64)
                end
                xTile = xTile + 1
                x = 0
            else
                collision = true
            end
        until collision == true
    elseif dir == 4 then
        repeat
            if xTile >= 0 and xTile < Game(9,23,10) then
                tile = pl:GetTileA(pl:GetTile(action, xTile, yTile))
                if ((tile.Attribute == 1 or tile.Attribute == 4) and tile.Type == pl.TileType.Single) or 
                --[[|]] (tile.Type == pl.TileType.Double and (tile.RectAttribute == 1 or tile.RectAttribute == 4) and (tile.Attribute == 1 or tile.Attribute == 4)) then 
                    collision = true
                elseif tile.Type == pl.TileType.Double and (tile.RectAttribute == 1 or tile.RectAttribute == 4) then -- double inside solid or death, outside clear
                    deltaDistance, collision = MoveXPointTile(x,y,tile.TileRect,4,true)
                    distance = distance + deltaDistance
                elseif tile.Type == pl.TileType.Double and (tile.Attribute == 1 or tile.Attribute == 4) then -- double outside solid or death
                    deltaDistance, collision = MoveXPointTile(x,y,tile.TileRect,4,false)
                    distance = distance + deltaDistance
                else -- all the other tiles
                    distance = distance + math.fmod(x,64) + 1
                end
                xTile = xTile - 1
                x = 63     
            else
                collision = true
            end
        until collision == true
    elseif dir == 2 then
        repeat
            if yTile >= 0 and yTile < Game(9,23,11) then
                tile = pl:GetTileA(pl:GetTile(action, xTile, yTile))
                if (tile.Attribute > 0 and tile.Type == pl.TileType.Single) or 
                --[[|]] (tile.Type == pl.TileType.Double and tile.RectAttribute > 0 and tile.Attribute > 0) then 
                    collision = true
                elseif tile.Type == pl.TileType.Double and tile.RectAttribute > 0 then -- double with outside clear, inside not clear
                    deltaDistance, collision = MoveYPointTile(x,y,tile.TileRect,2,true)
                    distance = distance + deltaDistance
                elseif tile.Type == pl.TileType.Double and tile.Attribute > 0 then -- double outside not clear
                    deltaDistance, collision = MoveXPointTile(x,y,tile.TileRect,2,false)
                    distance = distance + deltaDistance
                else -- all the other tiles
                    distance = distance + 64 - math.fmod(y,64)
                end
                yTile = yTile + 1
                y = 0
            else
                collision = true
            end
        until collision == true
    elseif dir == 8 then
        repeat
            if yTile >= 0 and yTile < Game(9,23,10) then
                tile = pl:GetTileA(pl:GetTile(action, xTile, yTile))
                if ((tile.Attribute == 1 or tile.Attribute == 4) and tile.Type == pl.TileType.Single) or 
                --[[|]] (tile.Type == pl.TileType.Double and (tile.RectAttribute == 1 or tile.RectAttribute == 4) and (tile.Attribute == 1 or tile.Attribute == 4)) then 
                    collision = true
                elseif tile.Type == pl.TileType.Double and (tile.RectAttribute == 1 or tile.RectAttribute == 4) then -- double inside solid or death, outside clear
                    deltaDistance, collision = MoveYPointTile(x,y,tile.TileRect,8,true)
                    distance = distance + deltaDistance
                elseif tile.Type == pl.TileType.Double and (tile.Attribute == 1 or tile.Attribute == 4) then -- double outside solid or death
                    deltaDistance, collision = MoveYPointTile(x,y,tile.TileRect,8,false)
                    distance = distance + deltaDistance
                else -- all the other tiles
                    distance = distance + math.fmod(y,64) + 1
                end 
                yTile = yTile - 1
                y = 63
            else
                collision = true
            end
        until collision == true
    end
    --
    return distance
end

function pl:PhysicsRect(x,y)
    return ffi.new("Rect", {pl:CollisionDistance(x,y,4), pl:CollisionDistance(x,y,8), pl:CollisionDistance(x,y,6), pl:CollisionDistance(x,y,2)})
end


local function YRayTile(ray, rect, xTile, inside)
    local _tileRect = RealTileRect(rect)
    local collision = false
    local distance;
    local tileW = pl:GetPlane(pl:MainPlane(), "TileWidth")
    local tileH = pl:GetPlane(pl:MainPlane(), "TileHeight")
    local x = xTile*tileW
    if inside == true then
        if ray.Left > x + _tileRect.Right or ray.Right < x + _tileRect.Left or math.abs(_tileRect.Top - _tileRect.Bottom) == 0 then
            collision = false
            distance = tileH - 1
        else
            collision = true
            distance = _tileRect.Top
        end
    else
        if _tileRect.Top > 0 then -- immediate contact
            collision = true
            distance = 0
        elseif (ray.Right > x + _tileRect.Right and _tileRect.Right < tileW - 1) or (ray.Left < x + _tileRect.Left and _tileRect.Left > 0) then
            collision = true
            distance = 0
        elseif _tileRect.Bottom < tileH - 1 then -- delayed contact
            collision = true
            distance = _tileRect.Bottom
        else -- no contact
            collision = false
            distance = tileH - 1
        end
    end
    return collision, distance
end

local function minInTable(t)
    if #t == 0 then return nil, nil end
    local key, value = 1, t[1]
    for i = 2, #t do
        if t[i] < value then
            key, value = i, t[i]
        end
    end
    return key,value
end

function pl:GetTileBelow(object)
    local action = pl:MainPlane()
    local tileW = pl:GetPlane(action, "TileWidth")
    local tileH = pl:GetPlane(action, "TileHeight")
    local ray = ffi.new("Rect", {object.X+object.MoveRect.Left, object.Y+object.MoveRect.Bottom, object.X+object.MoveRect.Right, object.Y+object.MoveRect.Bottom})
    local xTileL = math.floor(ray.Left/tileW)
    local xTileR = math.floor(ray.Right/tileW)
    local yTile = math.floor(ray.Bottom/tileH)
    local tile = ffi.new("int")
    local tileA = ffi.new("TileA*");
    local distance;
    local collision = false
    local firstCollision = false
    -- repeatedly check for collision all the tiles on the next ray line:
    repeat
        for xTile = xTileL, xTileR do
            tile = pl:GetTile(action, xTile, yTile)
            if tile >= 0 then
                tileA = pl:GetTileA(tile)
                if (tileA.Attribute > 0 and tileA.Type == pl.TileType.Single) or 
                --[[|]] (tileA.Type == pl.TileType.Double and tileA.RectAttribute > 0 and tileA.Attribute > 0) then -- single or double not clear
                    collision = true 
                elseif tileA.Type == pl.TileType.Double and tileA.RectAttribute > 0 then -- double inside not clear, outside clear
                    firstCollision, distance = YRayTile(ray, tileA.TileRect, xTile, true)
                    if firstCollision == true then 
                        collision = true 
                    end
                elseif tileA.Type == pl.TileType.Double and tileA.Attribute > 0 then -- double outside not clear, inside clear
                    firstCollision, distance = YRayTile(ray, tileA.TileRect, xTile, false)
                    if firstCollision == true then 
                        collision = true 
                    end
                end
            elseif tile == -3 then
                collision = true break
            end
        end
        yTile = yTile + 1
    until collision == true
    yTile = yTile - 1
    -- get the list of all the tiles on the collision line:
    local finalTiles, finalDistances = {}, {}
    for xTile = xTileL, xTileR do
        tile = pl:GetTile(action, xTile, yTile)
        if tile >= 0 then
            table.insert(finalTiles, tile)
            tileA = pl:GetTileA(tile)
            if (tileA.Attribute > 0 and tileA.Type == pl.TileType.Single) or 
                --[[|]] (tileA.Type == pl.TileType.Double and tileA.RectAttribute > 0 and tileA.Attribute > 0) then -- single or double not clear
                ray.Bottom = yTile*tileH
                table.insert(finalDistances, ray.Bottom)
            elseif tileA.Type == pl.TileType.Double and tileA.RectAttribute > 0 then -- double inside not clear, outside clear
                collision, distance = YRayTile(ray, tileA.TileRect, xTile, true)
                ray.Bottom = yTile*tileH + distance - 1
                table.insert(finalDistances, ray.Bottom)
            elseif tileA.Type == pl.TileType.Double and tileA.Attribute > 0 then -- double outside not clear, inside clear
                --TextOut("Tile.Left = "..tileA.TileRect.Left+xTile*tileW..", Ray.Left = "..ray.Left..", Tile.Right = "..tileA.TileRect.Right+xTile*tileW..", Ray.Right = "..ray.Right)
                collision, distance = YRayTile(ray, tileA.TileRect, xTile, false)
                ray.Bottom = yTile*tileH + distance
                table.insert(finalDistances, ray.Bottom)
            end
        elseif tile == -3 then -- out of plane borders
            return {ID = -3, Distance = yTile*tileH - ray.Top}
        end
    end
    -- get the final tile:
    local k,lowestY = minInTable(finalDistances)
    --TextOut("ID: "..finalTiles[k]..", Distance: "..lowestY - ray.Top..", Ray.Bottom: "..lowestY..", Ray.Top: "..ray.Top)
    return {ID = finalTiles[k], Distance = lowestY - ray.Top}
end

------------------------------------------------------------------------------------------ SCREEN ------------------------------------------------------------------------------------------

function pl:OnScreen()
    return ffi.new("Rect", {Game(9,23,16), Game(9,23,17), Game(9,23,18), Game(9,23,19)})
end

return pl
